'''
.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

Active learning pairwise search engines
=======================================
.. autofunction:: similar
.. autofunction:: dissimilar
.. autoclass:: PairwiseFeatureLearner
'''
from __future__ import absolute_import, division, print_function

import collections
from itertools import ifilter, imap, islice
import logging
import operator
import math
import re

from scipy.spatial.distance import cosine
from sklearn.feature_extraction import DictVectorizer
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.linear_model import LogisticRegression

from dossier.fc import StringCounter
from dossier.label import CorefValue, Label
from dossier.web import engine_index_scan, streaming_sample
from dossier.models.folder import Folders
from dossier.models.soft_selectors import find_soft_selectors

logger = logging.getLogger(__name__)

itget = operator.itemgetter


def similar(store, label_store):
    '''An active learning search engine that returns similar results.

    This satisfies the :class:`dossier.web.SearchEngine` interface.

    In addition to the standard query inputs of ``content_id``,
    ``filter_pred``, and ``limit``, this search engine has optional
    parameters ``canopy_limit`` and ``label_limit`` that restrict
    the number of candidates loaded for in-memory ranking and the
    number of labels used for model training, respectively.
    '''
    return create_search_engine(store, label_store, similar=True)


def dissimilar(store, label_store):
    '''An active learning search engine that returns dissimilar results.

    This satisfies the :class:`dossier.web.SearchEngine` interface.

    In addition to the standard query inputs of ``content_id``,
    ``filter_pred``, and ``limit``, this search engine has optional
    parameters ``canopy_limit`` and ``label_limit`` that restrict
    the number of candidates loaded for in-memory ranking and the
    number of labels used for model training, respectively.
    '''
    return create_search_engine(store, label_store, similar=False)


def create_search_engine(store, label_store, similar=True):
    # N.B. `limit` caps the final results returned by the ranker.
    # This is distinct from `canopy_limit` and `label_limit`, where
    # they are used to keep resource use in check.
    def _(content_id, filter_pred, limit,
          canopy_limit=100, label_limit=100, subtopic_id=None,
          **kwargs):
        '''Creates an active learning search engine.

        This is a helper function meant to satisfy the interface of
        a search engine with a particular active learning model.
        '''
        # Are these maximums good enough? No idea. ---AG
        canopy_limit = str_to_max_int(canopy_limit, 1000)
        label_limit = str_to_max_int(label_limit, 1000)

        learner = PairwiseFeatureLearner(
            store, label_store, content_id,
            subtopic_id=subtopic_id,
            canopy_limit=canopy_limit, label_limit=label_limit)
        try:
            candidate_probs = sorted(
                learner.probabilities(), reverse=similar, key=itget(1))
        except InsufficientTrainingData:
            logger.info('Falling back to plain index scan...')
            index_scan = engine_index_scan(store)
            engine_result = index_scan(content_id, filter_pred, limit)
            return add_facets(add_soft_selectors(engine_result, **kwargs))

        ranked = ifilter(lambda t: filter_pred(t[0]), candidate_probs)
        results = imap(lambda ((cid, fc), p): learner.as_result(cid, fc, p),
                       ranked)
        top_results = list(islice(results, limit))
        return add_facets(add_soft_selectors({'results': top_results}, **kwargs))
    return _


def add_soft_selectors(engine_result, **kwargs):
    results = engine_result['results']
    ids_and_clean, fcs = [], {}
    for r in results:
        ids_and_clean.append((r[0], r[1].get(u'meta_clean_visible', '')))
        fcs[r[0]] = r[1]
    suggestions = find_soft_selectors(ids_and_clean, **kwargs)
    for s in suggestions:
        for hit in s['hits']:
            hit['title'] = get_title(fcs[hit['content_id']],
                                     default=s['phrase'])
    return dict(engine_result, **{'suggestions': suggestions})


def add_facets(engine_result, facet_features=None):
    '''construct a new result payload with `facets` added as a new
    top-level property that carries a mapping from unicode strings to
    lists of content_ids.  The `facet_features` lists the names of
    :class:`~dossier.fc.StringCounter` features in each result
    :class:`~dossier.fc.FeatureCollection` to harvest phrases for the
    facets list.  If not specified, `facet_features` defaults to
    `bowNP_sip`.

    The remainder of the facet logic is handled in the UI.

    '''
    if facet_features is None:
        facet_features = ['bowNP_sip']
    phrases = collections.defaultdict(list)
    for r in engine_result['results']:
        cid = r[0]
        fc = r[1]
        for fname in facet_features:
            for phrase in fc.get(u'bowNP_sip', []):
                phrases[phrase].append(cid)
    return dict(engine_result, **{'facets': dict(phrases)})


def get_title(fc, default=None):
    title = fc.get(u'title')
    if title is None:
        m = re.search('<title>(.*?)</title>', fc.get(u'meta_clean_html', ''))
        if m is not None:
            title = m.group(1)
    return title or default


class InsufficientTrainingData(Exception):
    pass


class PairwiseFeatureLearner(object):
    '''A pairwise active learning model.

    This active learning model applies
    :class:`~sklearn.linear_model.LogisticRegression` on-the-fly
    as a user (or simulated user) interacts with content
    via the web services provided by :mod:`dossier.web`.

    This reads :class:`~dossier.label.Label` objects from
    :class:`~dossier.label.LabelStore` and provides predictions of
    pairwise equivalence, which can be used for coreference resolution,
    clustering, and ranking.

    .. automethod:: dossier.models.PairwiseFeatureLearner.__init__
    .. automethod:: dossier.models.PairwiseFeatureLearner.probabilities
    '''
    def __init__(self, store, label_store, content_id, subtopic_id=None,
                 canopy_limit=None, label_limit=None):
        '''Build a new model.

        :param store: A store of feature collections.
        :type store: :class:`dossier.store.Store`
        :param label_store: A store of labels (ground truth data).
        :type label_store: :class:`dossier.label.LabelStore`
        :param str content_id: The query content id (which should correspond
                               to a feature collection in the ``store``).
                               If it doesn't, no results are returned.
        :param int canopy_limit: A limit on the number of results to return
                                 in the canopy (the initial index scan).
                                 This is meant to be a mechanism for resource
                                 control.
        :param int label_limit: A limit on the number of labels to use in
                                training. This is meant to be a mechanism for
                                resource control.
        '''
        self.store = store
        self.label_store = label_store
        self.folders = Folders(store.kvl)
        self.query_content_id = content_id
        self.query_subtopic_id = subtopic_id
        self.query_fc = None
        self.canopy_limit = canopy_limit
        self.label_limit = label_limit

    def as_result(self, cid, fc, p):
        fnames = sorted(set(self.query_fc.keys()).intersection(fc.keys()))
        intermediates = {n: {'kernel': 'cosine',
                             'feature1': n,
                             'feature2': n,
                             'kernel_value': None,
                             'weight': None,
                             'common_feature_values': []} for n in fnames}
        for n in fnames:
            intermediates[n]['weight'] = self.feature_weights.get(n)
        for n, qfeat, cfeat in ((n, self.query_fc[n], fc[n]) for n in fnames):
            if not isinstance(qfeat, StringCounter) \
                    or not isinstance(cfeat, StringCounter):
                continue
            vals = set(qfeat.keys()).intersection(cfeat.keys())
            intermediates[n]['common_feature_values'] = \
                sorted(filter(None, vals))

            all_vals = sorted(set(qfeat.keys()).union(cfeat.keys()))
            if len(all_vals) > 0:
                qcounts = [qfeat.get(v, 0) for v in all_vals]
                ccounts = [cfeat.get(v, 0) for v in all_vals]
                sim = cosine(qcounts, ccounts)
                if not math.isnan(sim):
                    intermediates[n]['kernel_value'] = sim
        return (cid, fc, {
            'probability': p,
            'intermediate_model_results': intermediates.values(),
        })

    def probabilities(self):
        '''Trains a model and predicts recommendations.

        If the query feature collection could not be found or if there
        is insufficient training data, an empty list is returned.

        Otherwise, a list of content objects (tuples of content
        id and feature collection) and probabilities is returned.
        The probability is generated from the model, and reflects
        confidence of the model that the corresponding content object
        is related to the query based on the ground truth data.

        On a large database, random samples are used for training, so
        this function is not deterministic.

        :rtype: ``list`` of
          ((``content_id``, :class:`dossier.fc.FeatureCollection`),
          probability)
        '''
        self.query_fc = self.store.get(self.query_content_id)
        if self.query_fc is None:
            logger.warning('Could not find FC for %s', self.query_content_id)
            return []

        # Try the canopy query before training, because if the canopy query
        # gives us nothing, then there's no point in the additional work.
        #
        # Possible optimization: If the canopy query yields fewer than N
        # results, then can we just return all of them? ---AG
        #
        # N.B Doing the canopy query first will cause things to be slower
        # when there is insufficient training data.
        candidates = self.canopy(limit=self.canopy_limit)
        if len(candidates) == 0:
            logger.info(
                'Could not find any candidates in a canopy query by '
                'scanning the following indexes: %s',
                ', '.join(self.store.index_names()))
            return []

        # Get labels from the database and translate them to the form
        # `[{-1, 1}, i, j]` where `i, j` are indices into the list
        # `content_objs`, which has type `[(content_id, FeatureCollection)]`.
        logger.info('Fetching labels...')
        labels = list(self.labels_from_query(limit=self.label_limit))
        logger.info('Fetching FCs from labels...')
        content_objs = self.content_objs_from_labels(labels)
        indexed_labels = labels_to_indexed_coref_values(content_objs, labels)

        logger.info('Training...')
        model = self.train(content_objs, indexed_labels)
        if model is None:
            logger.info(
                'Could not train model: insufficient training data. '
                '(query content id: %s)', self.query_content_id)
            raise InsufficientTrainingData

        feature_names, classifier, transformer = model
        return zip(candidates, self.classify(
            feature_names, classifier, transformer, candidates))

    def train(self, content_objs, idx_labels):
        '''Trains and returns a model using sklearn.

        If there are new labels to add, they can be added, returns an
        sklearn model which can be used for prediction and getting
        features.

        This method may return ``None`` if there is insufficient
        training data to produce a model.

        :param labels: Ground truth data.
        :type labels: list of ``({-1, 1}, index1, index2)``.
        '''
        # We have insufficient training data when there is only one or
        # fewer classes of labels.
        if len({lab[0] for lab in idx_labels}) <= 1:
            return None

        fcs = [fc for _, fc in content_objs]
        feature_names = vectorizable_features(fcs)
        dis = dissimilarities(feature_names, fcs)

        phi_dicts, labels = [], []  # lists are in correspondence
        for coref_value, i, j in idx_labels:
            # i, j are indices into the list `fcs`
            labels.append(coref_value)  # either -1 or 1
            phi_dict = {name: dis[name][i,j] for name in feature_names}
            phi_dicts.append(phi_dict)

        vec = dict_vector()
        training_data = vec.fit_transform(phi_dicts)

        model = LogisticRegression(class_weight='auto', penalty='l1')
        model.fit(training_data, labels)
        self.feature_weights = {name: model.coef_[0][i]
                                for i, name in enumerate(feature_names)}
        return feature_names, model, vec

    def classify(self, feature_names, classifier, transformer, candidates):
        '''Returns ``[probability]`` in correspondence with
        ``candidates``.

        Where each ``probability`` corresponds to the probability that
        the corresponding candidate is classified with a positive label
        given the training data.

        The list returned is in correspondence with the list of
        candidates given.

        N.B. The contract of this method should be simplified by
        bundling ``feature_names``, ``classifier`` and ``transformer``
        into one thing known as "the model." ---AG
        '''
        dis = {}
        for name in feature_names:
            vec = dict_vector()
            query = vec.fit_transform([get_feat(self.query_fc, name)])
            cans = vec.transform(get_feat(fc, name) for _, fc in candidates)
            dis[name] = 1 - pairwise_distances(
                cans, query, metric='cosine', n_jobs=1)[:,0]

        # in correspondence with `candidates`
        phi_dicts = transformer.transform(
            [{name: dis[name][i] for name in feature_names}
             for i in xrange(len(candidates))])
        return classifier.predict_proba(phi_dicts)[:,1]

    def canopy(self, limit=None):
        ids = streaming_sample(self.canopy_ids(limit_hint=hard_limit(limit)),
                               limit, hard_limit(limit))
        # I don't think it ever makes sense to include the query
        # as part of the candidate set.
        return filter(lambda (_, fc): fc is not None, self.store.get_many(ids))

    def canopy_ids(self, limit_hint=None):
        limit_hint = limit_hint or 1000
        # TODO: It seems like this should pre-emptively discard content
        # ids that have already participated in a *direct* label with
        # the query. But I think this is a premature optimization since
        # the filtering functions will take care of it. (This optimization
        # would mean fewer kernel computations.)
        blacklist = {self.query_content_id}
        cids = set()

        # OK, so it turns out that a naive index scan is pretty inflexible and
        # arbitrary. The issue is that in a big enough data set, the first
        # index scan will probably exhaust all of our result set, which
        # means result sets will never see any variety.
        #
        # Instead, we'll try to sample from each index in small batch sizes.
        # This is a heuristic; not a principled approach. ---AG
        index_names = self.store.index_names()
        batch_size = limit_hint / 10
        progress = {}  # idx, name |--> last end
        # When `progress` is empty, the following loop will terminate.
        # An index is removed from `progress` when it no longer produces
        # results.
        for idx_name in index_names:
            feat = self.query_fc.get(idx_name)
            if isinstance(feat, StringCounter):
                for name in feat:
                    if len(name) > 0:
                        progress[(idx_name, name)] = 0

        logger.info('starting index scan (query content id: %s)',
                    self.query_content_id)
        while len(progress) > 0:
            for idx_name in index_names:
                for name in self.query_fc.get(idx_name, []):
                    key = (idx_name, name)
                    if key not in progress:
                        continue
                    logger.info('[StringCounter index: %s] scanning for "%s"',
                                idx_name, name)
                    scanner = self.store.index_scan(idx_name, name)
                    progressed = 0
                    for cid in islice(scanner, progress[key], None):
                        if progressed >= batch_size:
                            break
                        if cid not in cids and cid not in blacklist:
                            cids.add(cid)
                            progressed += 1
                            yield cid
                    if progressed == 0:
                        progress.pop(key)
                    else:
                        progress[key] += progressed

    def labels_from_query(self, limit=None):
        '''ContentId -> [Label]'''
        return self.infer_subtopic_labels(limit=limit)

    def infer_subtopic_labels(self, limit=None):
        # The basic idea here is to aggressively gather truth data while
        # avoiding cross contamination with other subfolders. Since our query
        # is a (content_id, subtopic_id), we can use subtopic connected
        # components to achieve this.

        # Short aliases.
        cid, subid = self.query_content_id, self.query_subtopic_id

        # For positive labels, the only thing we can do is traverse the
        # subtopic connected component.
        # Don't impose a hard limit on positive labels. (There are probably
        # very few of them.)
        logger.info('Inferring positive labels for: %r', (cid, subid))
        pos_labels = (self.label_store.expand((cid, subid))
                      + list(self.positive_subtopic_labels()))
        logger.info('Inferring negative labels for: %r', (cid, subid))
        neg_labels = self.negative_subtopic_labels()

        pos_sample = streaming_sample(
            pos_labels, limit, limit=hard_limit(limit))
        neg_sample = streaming_sample(
            neg_labels, limit, limit=hard_limit(limit))
        print('-' * 79)
        print('POSITIVES\n', '\n'.join(map(repr, pos_sample)), '\n')
        print('-' * 79)
        print('NEGATIVES\n', '\n'.join(map(repr, neg_sample)))
        print('-' * 79)
        return pos_sample + neg_sample

    def positive_subtopic_labels(self):
        cid, subid = self.query_content_id, self.query_subtopic_id
        subfolders = list(self.folders.parent_subfolders((cid, subid)))

        for fid, subfolder_id in subfolders:
            for cid2, subid2 in self.folders.items(fid, subfolder_id):
                # Since this item is in the same folder as our query, we
                # consider it a positive example. But there's no explicit
                # label for it, so manufacture one.
                #
                # TODO: Fix annotator id here. (We need to push annotator
                # information down into the search engine; the rest is
                # trivial.) ---AG
                yield Label(cid, cid2,
                            Folders.DEFAULT_ANNOTATOR_ID,
                            CorefValue.Positive,
                            subid, subid2)

                # Sometimes the user will directly attach a positive label
                # to an item in the folder. This will grab those.
                for lab in self.label_store.directly_connected(cid2):
                    if lab.value == CorefValue.Positive \
                            and lab.subtopic_for(cid2) == subid2:
                        yield lab

    def negative_subtopic_labels(self):
        cid, subid = self.query_content_id, self.query_subtopic_id
        subfolders = list(self.folders.parent_subfolders((cid, subid)))

        # Find any directly connected negative labels to any item in the
        # containing subfolder.
        for fid, subfolder_id in subfolders:
            for cid2, subid2 in self.folders.items(fid, subfolder_id):
                for lab in self.label_store.directly_connected(cid2):
                    if lab.value == CorefValue.Negative \
                            and lab.subtopic_for(cid2) == subid2:
                        yield lab

        # Find all items in subfolders other than the subfolder that contains
        # (cid, subid) and add negative labels. Stay inside the folder (topic)
        # for now though.
        #
        # It's possible that `(cid, subid)` are in more than one subfolder,
        # but in SortingDesk, `subid` is usually some kind of offset or hash,
        # so it's probably very unlikely. In any case, if it is in more than
        # one subfolder, then it's a user error and we just have to hope that
        # the model figures it out.
        in_fids = set()
        for fid, subfolder_id in subfolders:
            in_fids.add(fid)
            for cousin_subid in self.folders.subfolders(fid):
                if cousin_subid == subfolder_id:
                    # You can't be a cousin to yourself!
                    continue
                for cid2, subid2 in self.folders.items(fid, cousin_subid):
                    # TODO: Fix annotator id here. (We need to push annotator
                    # information down into the search engine; the rest is
                    # trivial.) ---AG
                    yield Label(cid, cid2,
                                Folders.DEFAULT_ANNOTATOR_ID,
                                CorefValue.Negative,
                                subid, subid2)

        # If we exhaust the above, then let's start adding negative labels with
        # other topics.
        for other_fid in self.folders.folders():
            if other_fid in in_fids:
                # The item was found in one of these folders above, so ignore
                # it here.
                continue
            # We're home free. Find every item in this folder and make a
            # negative label for each.
            for other_subid in self.folders.subfolders(other_fid):
                for cid2, subid2 in self.folders.items(other_fid, other_subid):
                    yield Label(cid, cid2,
                                Folders.DEFAULT_ANNOTATOR_ID,
                                CorefValue.Negative,
                                subid, subid2)

    def content_objs_from_labels(self, labels):
        '''[Label] -> [(content_id, FeatureCollection)]'''
        is_mapping = lambda obj: isinstance(obj, collections.Mapping)
        def is_valid_fc((cid, fc)):
            if fc is None:
                return False
            if sum(1 for name in fc if is_mapping(fc[name])) == 0:
                return False
            return True

        ids = set()
        for lab in labels:
            ids.add(lab.content_id1)
            ids.add(lab.content_id2)
        return list(ifilter(is_valid_fc, self.store.get_many(ids)))


def labels_to_indexed_coref_values(content_objs, labels):
    '''[(content_id, FeatureCollection)] -> [Label] -> [({-1,1}, i, j)]

    where 0 <= i, j < len(content_objs).
    '''
    cids_to_idx = {}
    for i, (content_id, _) in enumerate(content_objs):
        cids_to_idx[content_id] = i
    idx = lambda cid: cids_to_idx[cid]
    labs = []
    for lab in labels:
        if lab.content_id1 in cids_to_idx and lab.content_id2 in cids_to_idx:
            labs.append((lab.value.value, idx(lab.content_id1), idx(lab.content_id2)))
    return labs


def vectorizable_features(fcs):
    '''Discovers the ordered set of vectorizable features in ``fcs``.

    Returns a list of feature names, sorted lexicographically.
    Feature names are only included if the corresponding
    features are vectorizable (i.e., they are an instance of
    :class:`collections.Mapping`).
    '''
    is_mapping = lambda obj: isinstance(obj, collections.Mapping)
    return sorted({name for fc in fcs for name in fc if is_mapping(fc[name])})


def dissimilarities(feature_names, fcs):
    '''Computes the pairwise dissimilarity matrices.

    This returns a dictionary mapping each name in ``feature_names``
    to a pairwise dissimilarities matrix. The dissimilaritiy scores
    correspond to ``1 - kernel`` between each feature of each
    pair of feature collections in ``fcs``.

    (The kernel used is currently fixed to ``cosine`` distance.)
    '''
    dis = {}
    for count, name in enumerate(feature_names, 1):
        logger.info('computing pairwise dissimilarity matrix '
                    'for %d of %d features (current feature: %s)',
                    count, len(feature_names), name)
        ## opportunity to use joblib is buried down inside call to
        ## pairwise_distances...
        # And it's fixed to `n_jobs=1` because running multiprocessing
        # inside py.test causes weird problems. It also doesn't seem like a
        # good idea to do it inside a web server either. ---AG
        dis[name] = 1 - pairwise_distances(
            dict_vector().fit_transform([get_feat(fc, name) for fc in fcs]),
            metric='cosine', n_jobs=1)
    return dis


def dict_vector():
    return DictVectorizer(sparse=False)


def hard_limit(limit):
    return limit if limit is None else (limit * 10)


def get_feat(fc, name):
    if len(fc[name]) == 0:
        return StringCounter({'': 0})
    else:
        return fc[name]


def str_to_max_int(s, maximum):
    try:
        return min(maximum, int(s))
    except (ValueError, TypeError):
        return maximum



'''
.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

Generate feature collections with your data
===========================================
This library ships with a command line program ``dossier.etl`` which
provides a rudimentary pipeline for transforming data from your database
to feature collections managed by :mod:`dossier.store`.

(Currently, ``dossier.etl`` is hard-coded to support a specific HBase
database, but it will be generalized as part of future work.)
'''
from __future__ import absolute_import, division, print_function

import abc
import logging
import time
import urllib

import gensim
import kvlayer
from dossier.store import Store
from streamcorpus import Chunk
from streamcorpus_pipeline.stages import Configured
from streamcorpus_pipeline._clean_visible import cleanse, make_clean_visible
from streamcorpus_pipeline._clean_html import make_clean_html

from dossier.fc import FeatureCollection, StringCounter
import dossier.models.features as features

logger = logging.getLogger(__name__)

class ETL(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def cids_and_fcs(self):
        raise NotImplementedError


class to_dossier_store(Configured):
    '''A :mod:`streamcorpus_pipeline` `writer` stage with one optional parameter:

    .. code-block:: yaml

        tfidf_path: path/to/tfidf.data

    '''
    config_name = 'to_dossier_store'
    default_config = {
        'tfidf_path': None
    }
    def __init__(self, *args, **kwargs):
        super(to_dossier_store, self).__init__(*args, **kwargs)
        kvl = kvlayer.client()
        self.store = Store(kvl)
        tfidf_path = self.config.get('tfidf_path')
        self.tfidf = gensim.models.TfidfModel.load(tfidf_path)

    def process(self, t_path, name_info, i_str):
        '''converts each :attr:`streamcorpus.StreamItem.body.clean_html` from
        `t_path` into a :class:`~dossier.fc.FeatureCollection` and saves it in
        a :class:`~dossier.store.Store` configured with the global `kvlayer`
        config.

        '''
        def cids_and_fcs():
            count = 0
            seen = set()
            for si in Chunk(t_path):
                clean_html = getattr(si.body, 'clean_html', '')
                if clean_html is None or len(clean_html.strip()) == 0:
                    logger.warn('dropping SI lacking clean_html: %r', si.abs_url)
                    continue
                fc = html_to_fc(
                    clean_html=si.body.clean_html.decode('utf-8'),
                    clean_visible=si.body.clean_visible.decode('utf-8'),
                    encoding='utf-8',
                    url=si.abs_url,
                    timestamp=si.stream_time.epoch_ticks,
                )
                add_sip_to_fc(fc, self.tfidf)
                content_id = str(fc['content_id'])
                if content_id in seen:
                    logger.warn('dropping duplicate content_id=%r', content_id)
                else:
                    seen.add(content_id)
                    yield content_id, fc
                    count += 1
            logger.info('saved %d FCs from %d SIs', count, len(seen))
        self.store.put(cids_and_fcs())
        ## interface spec of streamcorpus_pipeline writers requires
        ## returning a list of locally generated paths.
        return []

    __call__ = process


def mk_content_id(key):
    return 'web|' + urllib.quote(key, safe='~')


def html_to_fc(html=None, clean_html=None, clean_visible=None, encoding=None, url=None, timestamp=None, other_features=None):
    '''`html` is expected to be a raw string received over the wire from a
    remote webserver, and `encoding`, if provided, is used to decode
    it.  Typically, encoding comes from the Content-Type header field.
    The :func:`~streamcorpus_pipeline._clean_html.make_clean_html`
    function handles character encodings.

    '''
    def add_feature(name, xs):
        if name not in fc:
            fc[name] = StringCounter()
        fc[name] += StringCounter(xs)

    timestamp = timestamp or int(time.time() * 1000)
    other_features = other_features or {}

    if clean_html is None:
        if html is not None:
            clean_html_utf8 = make_clean_html(html, encoding=encoding)
            clean_html = clean_html_utf8.decode('utf-8')
        else:
            clean_html_utf8 = u''
            clean_html = u''
    else:
        clean_html_utf8 = u''

    if clean_visible is None or len(clean_visible) == 0:
        clean_visible = make_clean_visible(clean_html_utf8).decode('utf-8')
    elif isinstance(clean_visible, str):
        clean_visible = clean_visible.decode('utf-8')

    fc = FeatureCollection()
    fc[u'meta_raw'] = html and uni(html, encoding) or u''
    fc[u'meta_clean_html'] = clean_html
    fc[u'meta_clean_visible'] = clean_visible
    fc[u'meta_timestamp'] = unicode(timestamp)

    url = url or u''

    fc[u'meta_url'] = uni(url)
    fc[u'content_id'] = unicode(mk_content_id(url))

    add_feature(u'phone', features.phones(clean_visible))
    add_feature(u'email', features.emails(clean_visible))
    add_feature(u'bowNP', features.noun_phrases(cleanse(clean_visible)))

    add_feature(u'image_url', features.image_urls(clean_html))
    add_feature(u'a_url', features.a_urls(clean_html))

    ## get parsed versions, extract usernames
    fc[u'img_url_path_dirs'] = features.path_dirs(fc[u'image_url'])
    fc[u'img_url_hostnames'] = features.host_names(fc[u'image_url'])
    fc[u'img_url_usernames'] = features.usernames(fc[u'image_url'])
    fc[u'a_url_path_dirs'] = features.path_dirs(fc[u'a_url'])
    fc[u'a_url_hostnames'] = features.host_names(fc[u'a_url'])
    fc[u'a_url_usernames'] = features.usernames(fc[u'a_url'])

    for feat_name, feat_val in other_features.iteritems():
        fc[feat_name] = feat_val

    return fc


def add_sip_to_fc(fc, tfidf, limit=40):
    '''add "bowNP_sip" to `fc` using `tfidf` data
    '''
    if 'bowNP' not in fc:
        return
    if tfidf is None:
        return
    sips = features.sip_noun_phrases(tfidf, fc['bowNP'].keys(), limit=limit)
    fc[u'bowNP_sip'] = StringCounter(sips)


def uni(s, encoding=None):
    # unicode string feat
    if not isinstance(s, unicode):
        try:
            return unicode(s, encoding)
        except:
            try:
                return unicode(s, 'utf-8')
            except UnicodeDecodeError:
                return unicode(s, 'latin-1')
    return s



# An unbelievably disgusting compatibility layer because I'm being rushed.
# ---AG

from __future__ import absolute_import, division, print_function

from collections import defaultdict

import dossier.web as web
import yakonfig


class Folders(web.Folders):
    DEFAULT_ANNOTATOR_ID = 'unknown'

    def __init__(self, *args, **kwargs):
        # A horrible hack to create a new `Folders` instance with config.
        try:
            config = yakonfig.get_global_config('dossier.folders')
            # For old configs.
            if 'prefix' in config:
                config['namespace'] = config.pop('prefix')
        except KeyError:
            config = {}
        super(Folders, self).__init__(*args, **dict(config, **kwargs))

    @staticmethod
    def name_to_id(v):
        return v

    @staticmethod
    def id_to_name(v):
        return v

    def folders(self, *args, **kwargs):
        return [it.name for it in self.list('/') if it.is_folder()]

    def subfolders(self, fid, *args, **kwargs):
        return [it.name for it in self.list(fid) if it.is_folder()]

    def items(self, fid, subid, *args, **kwargs):
        vals = []
        for it in self.list(fid + '/' + subid):
            if '@' in it.name:
                vals.append(tuple(map(lambda s: s.encode('utf-8'),
                                      it.name.split('@'))))
            else:
                vals.append((it.name.encode('utf-8'), None))
        return vals

    def grouped_items(self, fid, subid, *args, **kwargs):
        d = defaultdict(list)
        for cid, subid in self.items(fid, subid):
            d[cid].append(subid)
        return d

    def parent_subfolders(self, ident, *args, **kwargs):
        cid, _ = normalize_ident(ident)

        # There's no index structure to find parents of one path component.
        # So we need to do an exhaustive search. Yay!
        vals = []
        for folder in list(self.list('/')):
            if not folder.is_folder():
                continue
            for subfolder in list(self.list(folder.path)):
                if not subfolder.is_folder():
                    continue
                for item in list(self.list(subfolder.path)):
                    if not item.is_item():
                        continue
                    if cid in item.name.split('@'):
                        vals.append((folder.name, subfolder.name))
        return vals


def normalize_ident(ident):
    '''Splits a generic identifier.

    If ``ident`` is a tuple, then ``(ident[0], ident[1])`` is returned.
    Otherwise, ``(ident[0], None)`` is returned.
    '''
    if isinstance(ident, tuple) and len(ident) == 2:
        return ident[0], ident[1]  # content_id, subtopic_id
    else:
        return ident, None



'''``dossier.models.features.basic`` provides simple transforms that
construct features.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

'''
from __future__ import absolute_import, division, print_function

from itertools import imap
import logging
import regex as re
import traceback

from bs4 import BeautifulSoup
import urlnorm

from urlparse import urlparse
from dossier.fc import StringCounter


logger = logging.getLogger(__name__)


REGEX_PHONES = [
    r'\d-?\d{3}-?\d{3}-?\d{4}',
    r'\d{3}-?\d{3}-?\d{4}',
    r'\d{3}-?\d{4}',
]
REGEX_PHONE = re.compile('|'.join(REGEX_PHONES))

# Be extremely liberal with what we consider an email address.
REGEX_EMAIL = re.compile(r'\b\S+@\S+\.\S+\b', re.IGNORECASE)


def phones(text):
    '''Returns list of phone numbers without punctuation.'''
    return imap(lambda m: m.group(0).replace('-', ''),
                REGEX_PHONE.finditer(text))


def emails(text):
    '''Returns list of phone numbers without punctuation.'''
    return imap(lambda m: m.group(0).lower(), REGEX_EMAIL.finditer(text))


def image_urls(html):
    soup = BeautifulSoup(html)
    for node in soup.find_all('img'):
        try:
            src = node['src']
        except KeyError:
            continue
        yield norm_url(src)


def a_urls(html):
    '''
    return normalized urls found in the 'a' tag
    '''
    soup = BeautifulSoup(html)
    for node in soup.find_all('a'):
        try:
            href = node['href']
        except KeyError:
            continue
        yield norm_url(href)


def host_names(urls):
    '''
    Takes a StringCounter of normalized URL and parses their hostnames

    N.B. this assumes that absolute URLs will begin with

    http://

    in order to accurately resolve the host name.
    Relative URLs will not have host names.
    '''
    host_names = StringCounter()
    for url in urls:
        host_names[urlparse(url).netloc] += urls[url]
    return host_names


def path_dirs(urls):
    '''
    Takes a StringCounter of normalized URL and parses them into
    a list of path directories. The file name is
    included in the path directory list.
    '''
    path_dirs = StringCounter()
    for url in urls:
        for path_dir in filter(None, urlparse(url).path.split('/')):
            path_dirs[path_dir] += urls[url]
    return path_dirs


path_prefixes = r'''user|users|Users|home|data/media|var/users|u01''' + \
                r'''|Documents and Settings|WINNT\\Profiles'''

username_re = re.compile(
    r'^((?P<drive>[A-Za-z]):)?(/|\\)(%s)(/|\\)(?P<username>[^/\\$%%]+)' % path_prefixes
)

def usernames(urls):
    '''Take an iterable of `urls` of normalized URL or file paths and
    attempt to extract usernames.  Returns a list.

    '''
    usernames = StringCounter()
    for url, count in urls.items():
        uparse = urlparse(url)
        path = uparse.path
        hostname = uparse.hostname
        m = username_re.match(path)
        if m:
            usernames[m.group('username')] += count
        elif hostname in ['twitter.com', 'www.facebook.com']:
            usernames[path.lstrip('/')] += count
    return usernames


def norm_url(url):
    url = uni(url).encode('utf-8')
    try:
        return urlnorm.norm(url)
    except urlnorm.InvalidUrl:
        # Happens when the URL is relative. Call path normalization directly.
        return urlnorm.norm_path('', url)
    except:
        traceback.print_exc()
    return None


def uni(s):
    if not isinstance(s, unicode):
        try:
            return s.decode('utf-8')
        except UnicodeDecodeError:
            return s.decode('latin-1')
    return s




from dossier.models.etl import html_to_fc


test_html = '''

    <body id="ViewAd">
  <!-- Google Tag Manager -->
  <noscript><iframe src="//www.googletagmanager.com/ns.html?id=GTM-5KCSP8"
  height="0" width="0" style="display:none;visibility:hidden"></iframe></noscript>
  <script>(function(w,d,s,l,i){w[l]=w[l]||[];w[l].push({'gtm.start': 
  new Date().getTime(),event:'gtm.js'});var f=d.getElementsByTagName(s)[0],
  j=d.createElement(s),dl=l!='dataLayer'?'&l='+l:'';j.async=true;j.src=
  '//www.googletagmanager.com/gtm.js?id='+i+dl;f.parentNode.insertBefore(j,f);
  })(window,document,'script','bpDataLayer','GTM-5KCSP8');</script>
  <!-- End Google Tag Manager -->
    <div id="tlHeader">
<div id="postAdButton">
<form name="formPost" id="formPost" action="http://posting.newyork.backpage.com/online/classifieds/PostAdPPI.html/nyc/newyork.backpage.com/" method="get">
      <input type="submit" value="Post Ad" class="button" id="postAdButton">
      <input type="hidden" name="u" value="nyc">
      <input type="hidden" name="serverName" value="newyork.backpage.com">
    </form>
        </div><!-- #postAdButton -->
          <span class="formSearchHideOnSmallScreens">
          <input type="text" size="15" name="keyword" value=" keyword" onFocus="if (document.formSearch.keyword.value == ' keyword') document.formSearch.keyword.value = ''; return true;" maxlength="100">
            <select name="section">
                  <option value="26197783">local places
                  <option value="4382">community
                  <option value="4378">buy/ sell/ trade
                  <option value="153676">automotive
                  <option value="4380">musician
                  <option value="4376">rentals
                  <option value="4375">real estate
                  <option value="4373">jobs
'''

def test_html_to_fc():
    
    fc = html_to_fc(test_html.decode('utf8'))
    assert 'Date' not in fc['meta_clean_html']



'''Algorithmically determine soft selector strings.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.
'''

from __future__ import absolute_import, division, print_function
import argparse
from collections import defaultdict
import logging
from operator import itemgetter
import string

import dblogger
from gensim import corpora
import many_stop_words
from nltk.tokenize import RegexpTokenizer
from nltk.util import ngrams
import regex as re
import streamcorpus
from streamcorpus_pipeline._clean_html import clean_html
from streamcorpus_pipeline._clean_visible import clean_visible
import yakonfig


logger = logging.getLogger(__name__)
stop_words = many_stop_words.get_stop_words()


def find_soft_selectors(ids_and_clean_visible, start_num_tokens='10',
                        max_num_tokens='20',
                        filter_punctuation='0',
                        **kwargs):
    '''External interface for dossier.models.soft_selectors.

    This at scans through `num_tokens` values between
    `start_num_tokens` and `max_num_tokens` and calls
    `find_soft_selectors_at_n` looking for results

    All of the params can be passed from URL parameters, in which
    case they can be strings and this function will type cast them
    appropriately.
    '''
    start_num_tokens = int(start_num_tokens)
    max_num_tokens = int(max_num_tokens)
    filter_punctuation = bool(int(filter_punctuation))

    if not ids_and_clean_visible:
        logger.info('find_soft_selectors called with no ids_and_clean_visible')
        return []

    current_results = [] ## results from current n
    previous_results = [] ## previous results from last n
    overall_results = [] ## overall results to return

    for num_tokens in range(start_num_tokens, max_num_tokens + 1):

        ## update this here
        previous_results = current_results

        results_at_n = find_soft_selectors_at_n(
            ids_and_clean_visible, num_tokens, filter_punctuation)
        if len(results_at_n) == 0:
            break

        best_score = results_at_n[0]['score']

        ## i.e. the initial condition is they all have the same score
        idx_at_second = len(results_at_n)

        for idx, result in enumerate(results_at_n):
            if result['score'] < best_score:
                idx_at_second = idx
                break

        current_results = results_at_n[0:idx_at_second]

        if num_tokens == 8:
            for r in results_at_n:
                logger.info('%s --- score: %d' % (r['phrase'], r['score']))

        if previous_results == []:
            logger.info('Previous results are empty. Continuing.')
            continue

        ## now, the main idea is to figure out if any strings from previous
        ## are substrings of those from current
        ## (with the scores fixed at the max for that subphrase).
        ## when they stop being substrings
        ## then those are completed phrases and should be returned as a result

        for prev_result in previous_results:

            is_subbed_and_same_score = False

            for curr_result in current_results:
                if prev_result['phrase'] in curr_result['phrase'] and \
                    prev_result['score'] == curr_result['score'] :

                    is_subbed_and_same_score = True
                    break

            if not is_subbed_and_same_score: ## then it's a honest result
                prev_result['n'] = num_tokens - 1

                overall_results.append(prev_result)

        if len(current_results) == 0:
            ## we got them all
            ## (we still had to collect the previous results)
            ## that's why this break comes after the previous for loop
            break


    ## also add results from current_results at final n
    for result in current_results:
        result['n'] = num_tokens
        overall_results.append(result)

    ## sort by score then by length
    overall_results.sort(key=itemgetter('score', 'n'), reverse=True)

    logger.info('OVERALL RESULTS: %d' % len(overall_results))
    # for idx, result in enumerate(overall_results):
        # logger.info('%d. %s --- score: %f , n = %d, hits=%d' %
            # (idx, result['phrase'], result['score'], result['n'], len(result['hits']))
        # )
    return overall_results

def find_soft_selectors_at_n(ids_and_clean_visible, num_tokens,
                             filter_punctuation):
    corpus_clean_visibles = map(itemgetter(1), ids_and_clean_visible)
    corpus_cids = map(itemgetter(0), ids_and_clean_visible)

    corpus_strings = make_ngram_corpus(
        corpus_clean_visibles, num_tokens, filter_punctuation)


    ## make dictionary
    dictionary = corpora.Dictionary(corpus_strings)

    ## make word vectors
    corpus = map(dictionary.doc2bow, corpus_strings)

    # ## train tfidf model
    # tfidf = models.TfidfModel(corpus)

    # ## transform coprus
    # corpus_tfidf = tfidf[corpus]

    ## ignore tfidf
    corpus_tfidf = corpus

    ## sum up tf-idf across the entire corpus
    corpus_total = defaultdict(int)
    inverted_index = defaultdict(set)
    for doc_idx, doc in enumerate(corpus_tfidf):
        for word_id, score in doc:
            ## uncomment when doing tf idf
            # corpus_total[word_id] += score
            ## this way we only count once per doc
            corpus_total[word_id] += 1

            inverted_index[word_id].add(corpus_cids[doc_idx])

    ## order the phrases by score across the documents
    corpus_ordered = sorted(
        corpus_total.items(), key=itemgetter(1), reverse=True)

    top_phrases = []
    for word_id, score in corpus_ordered:
        top_phrases.append({
            'score': score,
            'phrase': dictionary[word_id],
            'hits': [{'content_id': cid, 'title': None}
                     for cid in inverted_index[word_id]],
        })
    return top_phrases


def make_ngram_corpus(corpus_clean_visibles, num_tokens, filter_punctuation,
                      zoning_rules=False):
    '''takes a list of clean_visible texts, such as from StreamItems or
    FCs, tokenizes all the texts, and constructs n-grams using
    `num_tokens` sized windows.

    `corpus_clean_visibles' -- list of unicode strings
    `num_tokens' --- the n of the n-grams
    `filter_punctuation' --- if True, punctuation is filtered
    '''
    ## TODO: generatlize this zoning code, so that it works on many
    ## sites in the HT domain; consider finishing streamcorpus-zoner
    ## to do this.
    if filter_punctuation:
        ## word tokenizer that removes punctuation
        tokenize = RegexpTokenizer(r'\w+').tokenize
        backpage_string = 'backpage'
        end_string = 'Poster'
    else:
        #tokenize = word_tokenize
        tokenize = lambda s: string.split(s)
        backpage_string = 'backpage.com'
        end_string = 'Poster\'s'

    corpus = list()
    for clean_vis in corpus_clean_visibles:
        ## crudely skip pages that have "error"
        if re.search(u'error', clean_vis, re.I & re.UNICODE):
            continue

        ## make tokens
        tokens = tokenize(clean_vis) ## already a unicode string

        if zoning_rules:
            ## filter out non backpage pages
            if backpage_string not in tokens:
                continue

            ## string that signals the beginning of the body
            try:
                idx0 = tokens.index('Reply')
            except:
                continue
            ## string that signals the end of the body
            try:
                idx1 = tokens.index(end_string)
            except:
                continue

            tokens = tokens[idx0:idx1]

        ## make ngrams, attach to make strings
        ngrams_strings = list()


        for ngram_tuple in ngrams(tokens, num_tokens):
            # ## attempt to remove unwanted phrases
            ## score with many_stop_words and drop bad tuples
            # stop_count = sum([int(bool(tok.lower() in stop_words))
            #                   for tok in ngram_tuple])
            # if stop_count > num_tokens / 1.5:
            #     continue

            ## remove ones with many repeated words
            if len(set(ngram_tuple)) < len(ngram_tuple) / 2:
                continue

            ## this adds ngrams for the current doc
            ngrams_strings.append(' '.join(ngram_tuple))


        ## this adds a list of all the ngrams from the current doc
        ## to the corpus list
        corpus.append(ngrams_strings)
    return corpus


def ids_and_clean_visible_from_streamcorpus_chunk_path(corpus_path):
    '''converts a streamcorpus.Chunk file into the structure that is
    passed by the search engine to find_soft_selectors

    '''
    ch = clean_html(clean_html.default_config)
    cv = clean_visible(clean_visible.default_config)
    ids_and_clean_visible = []
    for si in streamcorpus.Chunk(path=corpus_path):
        if not si.body.clean_visible:
            ## attempt to make clean_visible
            if not si.body.raw:
                logger.critical('no raw content, so skipping: %r', si.abs_url)
                continue
            abs_url = si.abs_url
            si = ch(si, {})
            if not si:
                logger.critical(
                    'failed to make clean_html, so skipping: %r', abs_url)
                continue
            si = cv(si, {})
            if not si or not si.body.clean_visible:
                logger.critical(
                    'failed to make clean_visible, so skipping: %r', abs_url)
                continue
        rec = (si.stream_id, si.body.clean_visible.decode('utf8'), {})
        ids_and_clean_visible.append(rec)
    return ids_and_clean_visible


def main():
    parser = argparse.ArgumentParser(
        'command line tool for debugging and development')
    parser.add_argument('corpus', help='path to a streamcorpus.Chunk file')
    parser.add_argument('-n', '--num-tokens', default=6, type=int,
                        help='the n of the ngrams; used as start_num_tokens '
                             'for scanning')
    parser.add_argument('--max-num-tokens', default=40, type=int,
                        help='maximum number of `n` in n-grams for scanning')
    parser.add_argument('--peak-score-delta', default=0.01, type=float,
                        help='delta in score values required between first '
                             'and second result to stop  scanning')
    parser.add_argument('--scan-window-size', default=False,
                        action='store_true',
                        help='if set, scans from the value of -n until it '
                             'finds a strongly peaked top value')
    parser.add_argument('--filter-punctuation', default=False,
                        action='store_true',
                        help='filter out punctuation; default is to not '
                             'filter punctuation')
    parser.add_argument('--show-ids', default=False, action='store_true',
                        help='show identifiers in diagnostic output')
    args = yakonfig.parse_args(parser, [yakonfig, dblogger])

    ## TODO: if we start needing to load FC chunk files (instead of SI
    ## chunk files), this might need to be told which kind of chunk it
    ## is loading, and we'll need a second function along the lines of
    ## ids_and_clean_visible_from_streamcorpus_chunk_path

    ## mimic the in-process interface:
    ids_and_clean_visible = ids_and_clean_visible_from_streamcorpus_chunk_path(
        args.corpus)
    logger.info('gathered %d texts', len(ids_and_clean_visible))

    def format_result(result):
        score, soft_selector_phrase, matching_texts = result
        return '%.6f\t%d texts say:\t%s\t%s' % \
            (score, len(matching_texts), soft_selector_phrase.encode('utf8'),
             args.show_ids and repr(matching_texts) or '')

    if args.scan_window_size:
        best = find_soft_selectors(
            ids_and_clean_visible,
            start_num_tokens=args.num_tokens,
            max_num_tokens=args.max_num_tokens,
            filtered_punctuation=args.filter_punctuation)
        if not best:
            print('failed to find a best result!')
        else:
            print('found a best result:')
            print('\n'.join(map(format_result, best)))

    else:
        results = find_soft_selectors_at_n(
            ids_and_clean_visible, args.num_tokens, args.filter_punctuation)
        print('\n'.join(map(format_result, results)))


if __name__ == '__main__':
    main()



from __future__ import absolute_import, division, print_function

from collections import defaultdict
import re
import traceback
import sys
import zlib

import cbor
import happybase
import happybase.hbase.ttypes as ttypes

from dossier.models.etl.interface import ETL, html_to_fc, mk_content_id
from dossier.fc import StringCounter


class Ads(ETL):
    def __init__(self, host, port, table_prefix=''):
        self.host = host
        self.port = port
        self.table_prefix = table_prefix
        self.last_key = None

    def get_conn(self):
        return happybase.Connection(host=self.host, port=self.port,
                                    table_prefix=self.table_prefix)

    def cids_and_fcs(self, mapper, start, end, limit=5):
        return mapper(row_to_content_obj,
                      self.get_artifact_rows(limit=limit,
                                             start_key=start, stop_key=end))

    def get_artifact_rows(self, limit=5, start_key=None, stop_key=None):
        while True:
            try:
                scanner = self.get_scanner(limit=limit,
                                           start_key=start_key,
                                           stop_key=stop_key)
                for key, data in scanner:
                    if key == self.last_key:
                        # Don't re-process troublesome keys.
                        continue
                    self.last_key = key
                    yield key, unpack_artifact_row(data)
                break
            except ttypes.IOError:
                # Die HBase, die!
                continue

    def get_scanner(self, limit=5, start_key=None, stop_key=None):
        if self.last_key is not None:
            start_key = self.last_key
        t = self.get_conn().table('artifact')
        return t.scan(row_start=start_key, row_stop=stop_key,
                      limit=limit, batch_size=20)


def row_to_content_obj(key_row):
    '''Returns ``FeatureCollection`` given an HBase artifact row.

    Note that the FC returned has a Unicode feature ``artifact_id``
    set to the row's key.
    '''
    key, row = key_row
    cid = mk_content_id(key.encode('utf-8'))
    response = row.get('response', {})

    other_bows = defaultdict(StringCounter)
    for attr, val in row.get('indices', []):
        other_bows[attr][val] += 1
    try:
        artifact_id = key
        if isinstance(artifact_id, str):
            artifact_id = unicode(artifact_id, 'utf-8')
        fc = html_to_fc(
            response.get('body', ''),
            url=row.get('url'), timestamp=row.get('timestamp'),
            other_features=dict(other_bows, **{'artifact_id': artifact_id}))
    except:
        fc = None
        print('Could not create FC for %s:' % cid, file=sys.stderr)
        print(traceback.format_exc(), file=sys.stderr)
    return cid, fc


def unpack_artifact_row(row):
    data = {
        'url': row['f:url'],
        'timestamp': int(row['f:timestamp']),
        'request': {
            'method': row['f:request.method'],
            'client': cbor.loads(zlib.decompress(row['f:request.client'])),
            'headers': cbor.loads(zlib.decompress(row['f:request.headers'])),
            'body': cbor.loads(zlib.decompress(row['f:request.body'])),
        },
        'response': {
            'status': row['f:response.status'],
            'server': {
                'hostname': row['f:response.server.hostname'],
                'address': row['f:response.server.address'],
            },
            'headers': cbor.loads(zlib.decompress(row['f:response.headers'])),
            'body': cbor.loads(zlib.decompress(row['f:response.body'])),
        },
        'indices': [],
    }
    for kk, vv in row.items():
        mm = re.match(r"^f:index\.(?P<key>.*)\.[0-9]+$", kk)
        if mm is not None:
            data['indices'].append((mm.group('key'), vv))
    return data



from __future__ import absolute_import, division, print_function

import re


javascript_words = map(lambda s: unicode(s).lower(), re.split(r'\s', '''
abstract  	else  	instanceof  	super
boolean  	enum  	int  	switch
break  	export  	interface  	synchronized
byte  	extends  	let  	this
case  	false  	long  	throw
catch  	final  	native  	throws
char  	finally  	new  	transient
class  	float  	null  	true
const  	for  	package  	try
continue  	function  	private  	typeof
debugger  	goto  	protected  	var
default  	if  	public  	void
delete  	implements  	return  	volatile
do  	import  	short  	while
double  	in  	static  	with
alert  	frames  	outerHeight
all  	frameRate  	outerWidth
anchor  	function  	packages
anchors  	getClass  	pageXOffset
area  	hasOwnProperty  	pageYOffset
Array  	hidden  	parent
assign  	history  	parseFloat
blur  	image  	parseInt
button  	images  	password
checkbox  	Infinity  	pkcs11
clearInterval  	isFinite  	plugin
clearTimeout  	isNaN  	prompt
clientInformation  	isPrototypeOf  	propertyIsEnum
close  	java  	prototype
closed  	JavaArray  	radio
confirm  	JavaClass  	reset
constructor  	JavaObject  	screenX
crypto  	JavaPackage  	screenY
Date  	innerHeight  	scroll
decodeURI  	innerWidth  	secure
decodeURIComponent  	layer  	select
defaultStatus  	layers  	self
document  	length  	setInterval
element  	link  	setTimeout
elements  	location  	status
embed  	Math  	String
embeds  	mimeTypes  	submit
encodeURI  	name  	taint
encodeURIComponent  	NaN  	text
escape  	navigate  	textarea
eval  	navigator  	top
event  	Number  	toString
fileUpload  	Object  	undefined
focus  	offscreenBuffering  	unescape
form  	open  	untaint
forms  	opener  	valueOf
frame  	option  	window
onbeforeunload  	ondragdrop  	onkeyup  	onmouseover
onblur  	onerror  	onload  	onmouseup
ondragdrop  	onfocus  	onmousedown  	onreset
onclick  	onkeydown  	onmousemove  	onsubmit
oncontextmenu  	onkeypress  	onmouseout  	onunload
'''))


def stopwords():
    return javascript_words[:]



#!/usr/bin/python
from __future__ import absolute_import, division, print_function

import argparse
from io import BytesIO
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
import sys
from urllib2 import urlopen
import re

from PIL import Image
import xlsxwriter

from dossier.store import Store
from dossier.models.folder import Folders
from dossier.models.subtopic import subtopics
import kvlayer
import yakonfig


class Factory(yakonfig.factory.AutoFactory):
    config_name = 'sortingdesk_report'
    kvlclient = property(lambda self: kvlayer.client())
    auto_config = lambda self: []


def main():
    p = argparse.ArgumentParser(
        description='SortingDesk report generation tool')
    p.add_argument('-c', '--config', required=True,
                   help='dossier stack YAML config file')
    p.add_argument('-o', '--output', required=True,
                   help='path to write Excel workbook file')
    p.add_argument(
        '-u', '--user', default='unknown', help='user name (default=ALL)')
    p.add_argument('folder', help='folder name')
    p.add_argument('subfolder', nargs='?', default=None,
                   help='subfolder name (default=ALL)')
    args = p.parse_args()

    config = yakonfig.set_default_config([kvlayer], filename=args.config)
    factory = Factory(config)
    store = factory.create(Store)

    # Instantiate and run report generator.
    folders = Folders(kvlayer.client())
    gen = ReportGenerator(store, folders, args.folder,
                          subfolder_name=args.subfolder, user=args.user)
    with open(args.output, 'wb+') as out:
        gen.run(out)


class ReportGenerator(object):
    '''Generates a report in Excel format.'''

    def __init__(self, store, folders, folder_name, subfolder_name=None, user='unknown'):
        '''Class constructor.

        :param folders: Reference to folder.Folders instance
        :param folder: folder name to generate report for
        :param subfolder: subfolder name; must be contained by folder and can
                          be None
        :param user: Generate report on data created by specified user.
        '''
        self.store = store
        self.folders = folders
        self.workbook = None
        self.formats = {}
        self.folder_name = folder_name
        self.subfolder_name = subfolder_name
        self.user = user

    @property
    def folder_id(self):
        return Folders.name_to_id(self.folder_name)

    @property
    def subfolder_id(self):
        return Folders.name_to_id(self.subfolder_name)

    def run(self, output):
        '''Generate the report to the given output.

        :param output: writable file-like object or file path
        '''
        # Ensure folder exists.
        if self.folder_id not in self.folders.folders(self.user):
            print("E: folder not found: %s" % self.folder_name,
                  file=sys.stderr)
            return

        # Create workbook.
        wb = self.workbook = xlsxwriter.Workbook(output)

        # Create the different styles used by this report generator.
        self.formats['title'] = wb.add_format({'font_size': '18',
                                               'bold': True})

        self.formats['default'] = wb.add_format({'align': 'top'})
        self.formats['bold'] = wb.add_format({'bold': True})

        self.formats['header'] = wb.add_format({
            'bold': True,
            'align': 'center',
            'valign': 'top',
            'font_size': '14',
            'font_color': '#506050',
            'bg_color': '#f5f5f5',
            'right': 1,
            'border_color': 'white'})

        self.formats['pre'] = wb.add_format({'font_name': 'Courier',
                                             'valign': 'top'})

        self.formats['link'] = wb.add_format({'valign': 'top',
                                              'font_color': 'blue',
                                              'underline': True})

        self.formats['type_text'] = wb.add_format({
            'font_color': '#BF8645',
            'valign': 'top',
            'align': 'center'})

        self.formats['type_image'] = wb.add_format({
            'font_color': '#84BF45',
            'valign': 'top',
            'align': 'center'})

        # Generate report for a specific subfolder or *all* subfolders of
        # self.folder .
        if self.subfolder_id is None:
            self._generate_report_all()
        else:
            self._generate_report_single(self.subfolder_id)

        # done and outta here
        self.workbook.close()

    def _generate_report_all(self):
        '''Generate report for all subfolders contained by self.folder_id.'''
        assert self.workbook is not None
        count = 0

        # Do all subfolders
        for sid in self.folders.subfolders(self.folder_id, self.user):
            count += 1
            self._generate_for_subfolder(sid)

        if count == 0:
            print("I: empty workbook created: no subfolders found")

    def _generate_report_single(self, sid):
        '''Generate report for subfolder given by sid .

        The main purpose of this method is to make sure the subfolder given by
        sid does indeed exist.  All real work is delegated to
        _generate_for_subfolder.

        :param sid: The subfolder id

        Private method.
        '''
        assert self.workbook is not None
        assert sid is not None

        # Ensure subfolder exists
        if not sid in self.folders.subfolders(self.folder_id, self.user):
            subfolder = Folders.id_to_name(sid)
            print("E: subfolder not found: %s" % subfolder, file=sys.stderr)
            return

        self._generate_for_subfolder(sid)

    def _generate_for_subfolder(self, sid):
        ''' Generate report for a subfolder.

        :param sid: The subfolder id; assumed valid
        '''
        # TODO: the following assumes subfolder names can be constructed from a
        # subfolder id, which might not be the case in the future.
        name = self._sanitise_sheetname(uni(Folders.id_to_name(sid)))
        ws = self.workbook.add_worksheet(name)
        fmt = self.formats
        ws.write("A1", "Dossier report", fmt['title'])
        ws.write("A2", "%s | %s" % (uni(self.folder_name), name))

        # Column dimensions
        ws.set_column('A:A', 37)
        ws.set_column('B:B', 37)
        ws.set_column('C:C', 37)
        ws.set_column('D:D', 8)
        ws.set_column('E:E', 30)
        ws.set_column('F:F', 37)

        # Header
        ws.write("A4", "Id", fmt['header'])
        ws.write("B4", "URL", fmt['header'])
        ws.write("C4", "Subtopic Id", fmt['header'])
        ws.write("D4", "Type", fmt['header'])
        ws.write("E4", "Content", fmt['header'])
        ws.write("F4", "Image URL", fmt['header'])

        # TODO: we probably want to wrap the following in a try-catch block, in
        # case the call to folders.subtopics fails.
        row = 4
        for i in subtopics(self.store, self.folders, self.folder_id, sid, self.user):
            Item.construct(self, i).generate_to(ws, row)
            row += 1

    def _sanitise_sheetname(self, sheetname):
        '''Sanitize worksheet names.

        The length of the sheetname is kept within 31 characters and any
        invalid chars are replaced by underscores.

        '''
        # Ensure that invalid characters are converted to underscore, whilst
        # making sure sheetname's length falls within 31 chars.
        return re.sub(r"[\[\]:*?/\\]", "_", sheetname[:31])


class Item(object):
    ''' Base class of concrete items ItemText and ItemImage. '''

    @staticmethod
    def construct(generator, subtopic):
        '''Method constructor of Item-derived classes.

        Given a subtopic tuple, this method attempts to construct an
        Item-derived class, currently either ItemText or ItemImage, from the
        subtopic's type, found in its 4th element.

        :param generator: Reference to the owning ReportGenerator instance
        :param subtopic: Tuple containing content_id, meta_url, subtopic_id,
        type and type-specific data.

        :returns An instantiated Item-derived class.

        '''
        type = subtopic[3]
        if type not in Item.constructors:
            raise LookupError(type)   # perhaps customise this exception?

        return Item.constructors[type](generator, subtopic)

    def __init__(self, generator, subtopic):
        self.generator = generator
        self.content_id, self.subtopic_id, self.meta_url, \
            type, self.data = subtopic[0:5]

    def generate_to(self, worksheet, row):
        fmt = self.generator.formats
        worksheet.write(row, 0, uni(self.content_id), fmt['pre'])
        worksheet.write(row, 1, uni(self.meta_url), fmt['link'])
        worksheet.write(row, 2, uni(self.subtopic_id), fmt['pre'])


class ItemImage(Item):
    ''' Represents an image item for the purpose of report generation. '''

    def __init__(self, generator, subtopic):
        '''Constructor.

        Delegates to base class constructor.

        '''
        super(ItemImage, self).__init__(generator, subtopic)

    def generate_to(self, worksheet, row):
        '''Generate row report.

        Generates a row report of the item represented by this instance and
        inserts it into a given worksheet at a specified row number.

        :param worksheet: Reference to a worksheet in which to insert row
        report.

        :param row: Row number.
        '''
        super(ItemImage, self).generate_to(worksheet, row)

        embedded = False
        fmt = self.generator.formats
        worksheet.write(row, 3, "image", fmt['type_image'])
        worksheet.write_url(row, 5, self.data[0], fmt['link'])

        try:
            if self.data:
                image = self.resize_image(StringIO(self.data[1]))
                worksheet.insert_image(row, 4, 'image',
                                       {'image_data': image})
                embedded = True
        except:
            # We probably wrongly ignoring the exception.  Should really at
            # least log it somewhere.
            pass

        # Attempt to retrieve image data from the image URL if image data not
        # present already above or something failed whilst recreating image
        # from base64 encoding.
        if not embedded:
            url = self.data[0]
            if url:
                image = self.resize_image(BytesIO(urlopen(url).read()))
                worksheet.insert_image(row, 4, url, {'image_data': image})
                embedded = True
            else:
                worksheet.write(row, 4, '<unavailable>')

        if embedded:
            worksheet.set_row(row, 40)

    def resize_image(self, data):
        '''Resize image if height over 50 pixels and convert to JPEG.

        Given a ByteIO or StringIO data input, this method ensures that the
        image is not over 50 pixels high.  If it is over 50 pixels high, the
        image is resized to precisely 50 pixels in height and the width is
        adjusted accordingly in keeping with the width/height ratio.  The image
        is always converted to JPEG to minimize any potentials issues while
        embedding the image in the Excel workbook.

        :param data: ByteIO or StringIO stream containing image data
        :returns Reference to a BytesIO instance containing resized image data.

        '''
        image = Image.open(data)
        stream_out = BytesIO()
        width, height = image.size[:]
        if height > 50:
            width = int(width * 50 / height)
            height = 50
            image = image.resize((width, 50))

        image.save(stream_out, format="JPEG", quality=100)

        stream_out.seek(0)
        return stream_out


class ItemText(Item):
    '''Represents a text snippet item for the purpose of report generation.'''

    def __init__(self, generator, subtopic):
        '''Constructor.

        Delegates to base class constructor.
        '''
        super(ItemText, self).__init__(generator, subtopic)

    def generate_to(self, worksheet, row):
        '''Generate row report.

        Generates a row report of the item represented by this instance and
        inserts it into a given worksheet at a specified row number.

        :param worksheet: Reference to a worksheet in which to insert row
        report.

        :param row: Row number.

        '''
        super(ItemText, self).generate_to(worksheet, row)

        fmt = self.generator.formats
        worksheet.write(row, 3, "text", fmt['type_text'])
        worksheet.write(row, 4, uni(self.data), fmt['default'])


class ItemManual(ItemText):
    pass


def uni(t):
    if t is None:
        return None
    if isinstance(t, unicode):
        return t
    return unicode(t, 'utf-8')


Item.constructors = {
    "text": ItemText,
    "image": ItemImage,
    "manual": ItemManual,
}
'''Map (dict) holding references to class constructors per item type.'''


if __name__ == '__main__':
    main()



'''``dossier.models.features.basic`` provides simple transforms that
construct features.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

'''
from dossier.fc import StringCounter
import dossier.models.features as features
import pytest


def test_extract_phones():
    txt = '''
Phone: 111-222-3333
Phone: 1112223333
Phone: 1-111-222-3333
Phone: 11112223333
Phone: 222-3333
Phone: 2223333
'''
    assert StringCounter(features.phones(txt)) == StringCounter({
        '1112223333': 2,
        '11112223333': 2,
        '2223333': 2,
    })


def test_a_urls():
    html = '''
<a href="http://ExAmPle.com/My Page.html">
<a href="http://example.com/My%20Page.html">
'''
    assert StringCounter(features.a_urls(html)) == StringCounter({
        'http://example.com/My Page.html': 2,
    })

def test_image_urls():
    html = '''
<img src="http://ExAmPle.com/My Image.jpg">
<img src="http://example.com/My%20Image.jpg">
'''
    assert StringCounter(features.image_urls(html)) == StringCounter({
        'http://example.com/My Image.jpg': 2,
    })


def test_extract_emails():
    txt = '''
email: abc@example.com
email: AbC@eXamPle.com
'''
    assert StringCounter(features.emails(txt)) == StringCounter({
        'abc@example.com': 2,
    })

def test_host_names():
    urls = StringCounter()
    urls['http://www.example.com/folder1'] = 3
    urls['http://www.example.com/folder2'] = 2
    urls['http://www.different.com/folder2'] = 7


    assert features.host_names(urls) == StringCounter({
        'www.example.com': 5,
        'www.different.com': 7,
    })

def test_path_dirs():
    urls = StringCounter()
    urls['http://www.example.com/folder1/folder3/index.html?source=dummy'] = 3
    urls['http://www.example.com/folder2/folder1'] = 2
    urls['http://www.different.com/folder2'] = 7


    assert features.path_dirs(urls) == StringCounter({
        'folder1': 5,
        'folder2': 9,
        'folder3': 3,
        'index.html': 3,
    })


example_usernames_from_paths = [
     (r'http://www.example.com/user/folder3/index.html?source=dummy', 'folder3', 3),
     (r'http://www.example.com/user/myaccount', 'myaccount', 2),
     (r'http://www.different.com/folder3', None, 4),
     (r'http://www.different.com/user/myaccount', 'myaccount', 7),
     (r'http://www.also.com/user', None, 23),
     (r'http://www.also2.com/user/user', 'user', 1),
     (r'http://frob.com/user/my_account/media/Dresses/hi.jpg', 'my_account', 1),
     (r'https://www.facebook.com/my_account', 'my_account', 1),
     (r'https://twitter.com/my_account', 'my_account', 1),
     (r'C:\WINNT\Profiles\myaccount%MyUserProfile%', 'myaccount', 3), # Microsoft Windows NT
     (r'C:\WINNT\Profiles\myaccount', 'myaccount', 3), # Microsoft Windows NT
     (r'd:\WINNT\Profiles\myaccount', 'myaccount', 3), # Microsoft Windows NT
     (r'X:\Documents and Settings\myaccount', 'myaccount', 8), # Microsoft Windows 2000, XP and 2003
     (r'C:\Users\myaccount', 'myaccount', 3), # Microsoft Windows Vista, 7 and 8
     (r'C:\Users\myaccount\dog', 'myaccount', 3), # Microsoft Windows Vista, 7 and 8
     (r'C:\Users\whg\Desktop\Plug\FastGui(LYT)\Shell\Release\Shell.pdb', 'whg', 2),
     (r'C:\Documents and Settings\whg\\Plug\FastGui(LYT)\Shell\Release\Shell.pdb', 'whg', 3),
     (r'C:\Users\whg\Desktop\Plug\FastGui(LYT)\Shell\Release\Shell.pdb', 'whg', 3),
     (r'/home/myaccount$HOME', 'myaccount', 5), # Unix-Based
     (r'/var/users/myaccount', 'myaccount', 3), # Unix-Derived
     (r'/u01/myaccount', 'myaccount', 3), # Unix-Derived
     (r'/user/myaccount', 'myaccount', 3), # Unix-Derived
     (r'/users/myaccount', 'myaccount', 3), # Unix-Derived
     (r'/var/users/myaccount', 'myaccount', 3), # Unix-Derived
     (r'/home/myaccount', 'myaccount', 3), # Linux / BSD (FHS)
     (r'/Users/my_account$HOME', 'my_account', 5), # Mac OS X
     (r'/Users/my_account', 'my_account', 5), # Mac OS X
     (r'/data/media/myaccount', 'myaccount', 5) # Android
     ]

@pytest.mark.parametrize(
    ('url_or_path', 'username', 'count'),
    example_usernames_from_paths
)
def test_usernames(url_or_path, username, count):
    urls = StringCounter()
    urls[url_or_path] += count

    if username is not None:
        results = features.usernames(urls)
        assert results == StringCounter({username: count})



from __future__ import absolute_import, division, print_function

import argparse
from itertools import imap
import multiprocessing
import sys

from gensim import models

from dossier.fc import FeatureCollectionChunk
from dossier.store import Store
from dossier.store.store import fc_dumps
from dossier.models.etl import Ads, Scrapy, add_sip_to_fc
import kvlayer
import yakonfig


HBASE_MAX_KEY_VALUE_SIZE = 10 * 1024 * 1024 - 100


def batch_progress(cids_and_fcs, add, limit=5, batch_size=100):
    def status(*args, **kwargs):
        kwargs['end'] = ''
        args = list(args)
        args[0] = '\033[2K\r' + args[0]
        print(*args, **kwargs)
        sys.stdout.flush()

    total = 'all' if limit is None else str(limit)
    status('0 of %s done' % total)
    batch = []
    last_cid = None
    for i, (cid, fc) in enumerate(cids_and_fcs, 1):
        if fc is None:
            continue
        # Since we can restart the scanner, we may end up regenerating
        # FCs for the same key in the same batch. This results in
        # undefined behavior in kvlayer.
        if not any(cid == cid2 for cid2, _ in batch):
            # HBase doesn't allow more than 10MB for any particular key/value.
            # This is supposedly configurable, but it's unclear how to
            # configure it with happybase.
            if data_size(cid, fc) <= HBASE_MAX_KEY_VALUE_SIZE:
                batch.append((cid, fc))

        if len(batch) >= batch_size:
            add(batch)
            last_cid = batch[-1][0]
            batch = []
        if i % 100 == 0:
            status('%d of %s done (last id: %r)'
                   % (i, 'all' if limit is None else str(limit), last_cid))
    if len(batch) > 0:
        add(batch)
    status('done')


def data_size(cid, fc):
    return len(cid) + len(fc_dumps(fc))


class App(yakonfig.cmd.ArgParseCmd):
    def __init__(self, *args, **kwargs):
        yakonfig.cmd.ArgParseCmd.__init__(self, *args, **kwargs)
        self._store = None
        self._chunk = None
        self.tfidf = None

    @property
    def store(self):
        if self._store is None:
            feature_indexes = None
            try:
                conf = yakonfig.get_global_config('dossier.store')
                feature_indexes = conf['feature_indexes']
            except KeyError:
                pass
            self._store = Store(kvlayer.client(),
                                feature_indexes=feature_indexes)
        return self._store

    def done(self):
        if self._chunk is not None:
            self._chunk.flush()

    def get_output_accumulator(self, output_path=None):
        if output_path is not None:
            self._chunk = FeatureCollectionChunk(path=output_path, mode='wb')
        def add(cids_and_fcs):
            if self.tfidf is not None:
                for _, fc in cids_and_fcs:
                    add_sip_to_fc(fc, self.tfidf)
            if output_path is not None:
                for _, fc in cids_and_fcs:
                    self._chunk.add(fc)
            else:
                self.store.put(cids_and_fcs)
        return add

    def get_mapper(self, args):
        cpus = getattr(args, 'processes', 1)
        if cpus == 1:
            return imap
        else:
            pool = multiprocessing.Pool(processes=cpus)
            return pool.imap

    def args_etl_ads(self, p):
        p.add_argument('--host', default='localhost')
        p.add_argument('--port', default=9090, type=int)
        p.add_argument('--table-prefix', default='')
        p.add_argument('--limit', default=None, type=int)
        p.add_argument('--batch-size', default=1000, type=int)
        p.add_argument('--start', default=None, type=str)
        p.add_argument('--stop', default=None, type=str)
        p.add_argument('-p', '--processes',
                       default=multiprocessing.cpu_count(), type=int)
        p.add_argument('-o', '--output', default=None)
        p.add_argument('--tfidf', default=None, type=str,
                       help='Path to TF-IDF background model. Can be '
                            'generated with the `dossier.etl tfidf` script.')

    def do_etl_ads(self, args):
        if args.tfidf is not None:
            self.tfidf = models.TfidfModel.load(args.tfidf)
        etl = Ads(args.host, args.port, table_prefix=args.table_prefix)
        gen = etl.cids_and_fcs(self.get_mapper(args), args.start, args.stop,
                               limit=args.limit)
        self.etl(args, etl, gen)

    def args_etl_scrapy(self, p):
        p.add_argument('-p', '--processes',
                       default=multiprocessing.cpu_count(), type=int)
        p.add_argument('--batch-size', default=1000, type=int)
        p.add_argument('--limit', default=None, type=int)
        p.add_argument('-o', '--output', default=None)
        p.add_argument('--url-prefix', default=None,
                       help='Override the URL prefix to use when fixing '
                            'relative URLs. When omitted, detect '
                            'automatically.')
        p.add_argument('--tfidf', default=None, type=str,
                       help='Path to TF-IDF background model. Can be '
                            'generated with the `dossier.etl tfidf` script.')
        p.add_argument('input',
                       help='Scrapy data. Only supports CSV format currently.')

    def do_etl_scrapy(self, args):
        if args.tfidf is not None:
            self.tfidf = models.TfidfModel.load(args.tfidf)

        url_prefix = args.url_prefix
        if url_prefix is None:
            url_prefix = Scrapy.detect_url_prefix(open(args.input))
            if url_prefix is not None:
                print('Auto-detected URL prefix:', url_prefix)
        etl = Scrapy(open(args.input), url_prefix=url_prefix)
        gen = etl.cids_and_fcs(self.get_mapper(args), limit=args.limit)
        self.etl(args, etl, gen)

    def etl(self, args, etl, gen):
        add = self.get_output_accumulator(args.output)
        try:
            batch_progress(gen, add, limit=args.limit,
                           batch_size=args.batch_size)
        finally:
            self.done()


def main():
    p = argparse.ArgumentParser(
        description='Utilities for generating FCs from artifacts.')
    app = App()
    app.add_arguments(p)
    args = yakonfig.parse_args(p, [kvlayer, yakonfig])
    app.main(args)


if __name__ == '__main__':
    main()



#!/usr/bin/python
from __future__ import absolute_import, division, print_function

import argparse
import sys

from dossier.label import LabelStore
from dossier.store import Store
import kvlayer
import yakonfig

from dossier.models.subtopic import Folders


class Factory(yakonfig.factory.AutoFactory):
    config_name = 'sortingdesk_report'
    kvlclient = property(lambda self: kvlayer.client())
    auto_config = lambda self: []


def main():
    p = argparse.ArgumentParser(
        description='SortingDesk report generation tool')
    p.add_argument('-c', '--config', required=True,
                   help='Dossier stack YAML config file')
    p.add_argument('folder_name', nargs='?', default=None, help='Folder name')
    p.add_argument('-u', '--user', default=None,help='user name (default=ALL)')
    args = p.parse_args()

    config = yakonfig.set_default_config([kvlayer], filename=args.config)
    factory = Factory(config)
    store = factory.create(Store)
    label_store = factory.create(LabelStore)
    folders = Folders(store, label_store)

    list_projects(folders, args.folder_name, args.user)


def list_projects(folders, folder = None, user = None):
    '''List all folders or all subfolders of a folder.

    If folder is provided, this method will output a list of subfolders
    contained by it.  Otherwise, a list of all top-level folders is produced.

    :param folders: reference to folder.Folders instance
    :param folder: folder name or None
    :param user: optional user name

    '''
    fid = None if folder is None else Folders.name_to_id(folder)

    # List all folders if none provided.
    if fid is None:
        for f in folders.folders(user):
            print(Folders.id_to_name(f))

        return

    # List subfolders of a specific folder
    try:
        for sid in folders.subfolders(fid, user):
            print(Folders.id_to_name(sid))
    except KeyError:
        print("E: folder not found: %s" %folder, file=sys.stderr)


if __name__ == '__main__':
    main()



'''dossier.models.tests provides a fixtures

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''
import pytest

from dossier.label import LabelStore
from dossier.store import Store
import kvlayer

@pytest.yield_fixture
def kvl():
    config = {
        'storage_type': 'local',
        'app_name': 'diffeo',
        'namespace': 'dossier.models.tests',
    }
    client = kvlayer.client(config)
    yield client
    client.close()


@pytest.yield_fixture
def store(kvl):
    client = Store(kvl, feature_indexes=[u'feature'])
    yield client
    client.delete_all()


@pytest.yield_fixture
def label_store(kvl):
    client = LabelStore(kvl)
    yield client
    client.delete_all()





from dossier.models.soft_selectors import make_ngram_corpus

def test_make_ngram_corpus():
    corpus_clean_visibles = ['''
Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture 
''',
'''
Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture 
''',
'''
Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture 
''',
'''
Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture 
''',
'''
Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture Enlarge Picture 
''',
'''
to decipher a hastily scribbled note
''',
'''
to decipher a hastily scribbled note
''',
'''
to decipher a hastily scribbled note
''',
'''
to decipher a hastily scribbled note
''',
'''
to decipher a hastily scribbled note
''',
]
    candidates = make_ngram_corpus(corpus_clean_visibles, 6, True)

    count = 0
    for c in candidates:
        if not c: continue
        assert c == ['to decipher a hastily scribbled note']
        count += 1

    assert count == 5
